@@ -4,13 +4,17 @@
 The LaunchPad manages the FireWorks database.
 """
 import datetime
+import os
 import time
+
+from pymongo.mongo_client import MongoClient
+from pymongo import DESCENDING
+
 from fireworks.core.fw_config import FWConfig
 from fireworks.utilities.fw_serializers import FWSerializable
-from pymongo.mongo_client import MongoClient
 from fireworks.core.firework import FireWork, Launch, Workflow
-from pymongo import DESCENDING
-from fireworks.utilities.fw_utilities import get_fw_logger, log_exception
+from fireworks.utilities.fw_utilities import get_fw_logger
+
 
 __author__ = 'Anubhav Jain'
 __copyright__ = 'Copyright 2013, The Materials Project'
@@ -36,7 +40,7 @@ class LaunchPad(FWSerializable):
     """
 
     def __init__(self, host='localhost', port=27017, name='fireworks', username=None, password=None,
-                 logdir=None, strm_lvl=None):
+                 logdir=None, strm_lvl=None, user_indices=None, wf_user_indices=None):
         """
         
         :param host:
@@ -46,6 +50,8 @@ class LaunchPad(FWSerializable):
         :param password:
         :param logdir:
         :param strm_lvl:
+        :param user_indices:
+        :param wf_user_indices:
         """
         self.host = host
         self.port = port
@@ -58,29 +64,46 @@ class LaunchPad(FWSerializable):
         self.strm_lvl = strm_lvl if strm_lvl else 'INFO'
         self.m_logger = get_fw_logger('launchpad', l_dir=self.logdir, stream_level=self.strm_lvl)
 
-        self.connection = MongoClient(host, port, j=True)
-        self.database = self.connection[name]
-        if username:
-            self.database.authenticate(username, password)
+        self.user_indices = user_indices if user_indices else []
+        self.wf_user_indices = wf_user_indices if wf_user_indices else []
 
-        self.fireworks = self.database.fireworks
-        self.launches = self.database.launches
-        self.fw_id_assigner = self.database.fw_id_assigner
-        self.workflows = self.database.workflows
+        # get connection
+        self.connection = MongoClient(host, port, j=True)
+        self.db = self.connection[name]
+        if username:
+            self.db.authenticate(username, password)
+
+        self.fireworks = self.db.fireworks
+        self.launches = self.db.launches
+        self.fw_id_assigner = self.db.fw_id_assigner
+        self.workflows = self.db.workflows
 
     def to_dict(self):
         """
         Note: usernames/passwords are exported as unencrypted Strings!
         """
         d = {'host': self.host, 'port': self.port, 'name': self.name, 'username': self.username,
-             'password': self.password, 'logdir': self.logdir, 'strm_lvl': self.strm_lvl}
+             'password': self.password, 'logdir': self.logdir, 'strm_lvl': self.strm_lvl,
+             'user_indices': self.user_indices, 'wf_user_indices': self.wf_user_indices}
         return d
 
     @classmethod
     def from_dict(cls, d):
         logdir = d.get('logdir', None)
         strm_lvl = d.get('strm_lvl', None)
-        return LaunchPad(d['host'], d['port'], d['name'], d['username'], d['password'], logdir, strm_lvl)
+        user_indices = d.get('user_indices', [])
+        wf_user_indices = d.get('wf_user_indices', [])
+        return LaunchPad(d['host'], d['port'], d['name'], d['username'], d['password'], logdir,
+                         strm_lvl, user_indices, wf_user_indices)
+
+    @classmethod
+    def auto_load(cls):
+        if FWConfig().LAUNCHPAD_LOC:
+            return LaunchPad.from_file(FWConfig().LAUNCHPAD_LOC)
+        elif FWConfig().CONFIG_FILE_DIR:
+            return LaunchPad.from_file(os.path.join(FWConfig().CONFIG_FILE_DIR, 'my_launchpad.yaml'))
+        return LaunchPad()
+
 
     def reset(self, password, require_password=True):
         """
@@ -128,7 +151,7 @@ class LaunchPad(FWSerializable):
             self.m_logger.debug('Sleeping for {} secs...'.format(maintain_interval))
             time.sleep(maintain_interval)
 
-    def add_wf(self, wf):
+    def add_wf(self, wf, reassign_all=True):
         """
 
         :param wf: a Workflow object.
@@ -138,7 +161,7 @@ class LaunchPad(FWSerializable):
             wf = Workflow.from_FireWork(wf)
 
         # insert the FireWorks and get back mapping of old to new ids
-        old_new = self._upsert_fws(wf.id_fw.values(), reassign_all=True)
+        old_new = self._upsert_fws(wf.id_fw.values(), reassign_all=reassign_all)
 
         # update the Workflow with the new ids
         wf._reassign_ids(old_new)
@@ -181,8 +204,15 @@ class LaunchPad(FWSerializable):
         launches = []
         for launch_id in fw_dict['launches']:
             launches.append(self.get_launch_by_id(launch_id).to_dict())
+
         fw_dict['launches'] = launches
 
+        archived_launches = []
+        for launch_id in fw_dict['archived_launches']:
+            archived_launches.append(self.get_launch_by_id(launch_id).to_dict())
+
+        fw_dict['archived_launches'] = archived_launches
+
         return FireWork.from_dict(fw_dict)
 
     def get_wf_by_fw_id(self, fw_id):
@@ -197,29 +227,52 @@ class LaunchPad(FWSerializable):
         fws = []
         for fw_id in links_dict['nodes']:
             fws.append(self.get_fw_by_id(fw_id))
-        links = Workflow.Links.from_dict(links_dict['links']).to_dict()  # necessary because Mongo no like int keys
 
-        return Workflow(fws, links, links_dict['metadata'])
+        return Workflow(fws, links_dict['links'], links_dict['name'], links_dict['metadata'])
 
-    def get_fw_ids(self, query=None, sort=False):
+    def get_fw_ids(self, query=None, sort=None, limit=0, count_only=False):
         """
         Return all the fw ids that match a query,
-        :param query: a dict representing a Mongo query
+        :param query: (dict) representing a Mongo query
+        :param sort: [(str,str)] sort argument in Pymongo format
+        :param limit: (int) limit the results
+        :param count_only: (bool) only return the count rather than explicit ids
         """
         fw_ids = []
         criteria = query if query else {}
-        sort = [("spec._priority", DESCENDING)] if sort else None
-        for fw in self.fireworks.find(criteria, {"fw_id": True}, sort=sort):
-            fw_ids.append(fw["fw_id"])
 
+        if count_only:
+            return self.fireworks.find(criteria, {"fw_id": True}, sort=sort).limit(limit).count()
+
+        for fw in self.fireworks.find(criteria, {"fw_id": True}, sort=sort).limit(limit):
+            fw_ids.append(fw["fw_id"])
         return fw_ids
 
-    def run_exists(self):
+    def get_wf_ids(self, query=None, sort=None, limit=0, count_only=False):
+        """
+        Return one fw id for all workflows that match a query,
+        :param query: (dict) representing a Mongo query
+        :param sort: [(str,str)] sort argument in Pymongo format
+        :param limit: (int) limit the results
+        :param count_only: (bool) only return the count rather than explicit ids
+        """
+        wf_ids = []
+        criteria = query if query else {}
+        if count_only:
+            return self.workflows.find(criteria, {"nodes": True}, sort=sort).limit(limit).count()
+
+        for fw in self.workflows.find(criteria, {"nodes": True}, sort=sort).limit(limit):
+            wf_ids.append(fw["nodes"][0])
+
+        return wf_ids
+
+    def run_exists(self, fworker=None):
         """
         Checks to see if the database contains any FireWorks that are ready to run
         :return: (T/F)
         """
-        return bool(self._get_a_fw_to_run(checkout=False))
+        q = fworker.query if fworker else {}
+        return bool(self._get_a_fw_to_run(query=q, checkout=False))
 
     def tuneup(self):
         self.m_logger.info('Performing db tune-up')
@@ -227,6 +280,10 @@ class LaunchPad(FWSerializable):
         self.m_logger.debug('Updating indices...')
         self.fireworks.ensure_index('fw_id', unique=True)
         self.fireworks.ensure_index('state')
+        self.fireworks.ensure_index('spec._category')
+        self.fireworks.ensure_index('created_on')
+        self.fireworks.ensure_index('name')
+
         self.launches.ensure_index('launch_id', unique=True)
         self.launches.ensure_index('state')
         self.launches.ensure_index('time_start')
@@ -234,20 +291,35 @@ class LaunchPad(FWSerializable):
         self.launches.ensure_index('host')
         self.launches.ensure_index('ip')
 
+        self.workflows.ensure_index('name')
+        self.workflows.ensure_index('created_on')
+        self.workflows.ensure_index('updated_on')
+
+        for idx in self.user_indices:
+            self.fireworks.ensure_index(idx)
+
+        for idx in self.wf_user_indices:
+            self.workflows.ensure_index(idx)
+
         self.m_logger.debug('Compacting database...')
         try:
-            self.database.command({'compact': 'fireworks'})
-            self.database.command({'compact': 'launches'})
+            self.db.command({'compact': 'fireworks'})
+            self.db.command({'compact': 'launches'})
         except:
-            log_exception(self.m_logger, 'Error while compacting database! (make sure your Mongo version is 2.0+)')
+            self.m_logger.debug('Database compaction failed (not critical)')
+
 
     def defuse_fw(self, fw_id):
         allowed_states = ['DEFUSED', 'WAITING', 'READY', 'FIZZLED']
-        return self.fireworks.find_and_modify({'fw_id': fw_id, 'state': {'$in': allowed_states}},
+        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': {'$in': allowed_states}},
                                               {'$set': {'state': 'DEFUSED'}})
 
+        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
+        return f
+
     def reignite_fw(self, fw_id):
-        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': 'DEFUSED'}, {'$set': {'state': 'WAITING'}})
+        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': 'DEFUSED'},
+                                           {'$set': {'state': 'WAITING'}})
         if f:
             self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
         return f
@@ -257,11 +329,28 @@ class LaunchPad(FWSerializable):
         for fw in wf.fws:
             self.defuse_fw(fw.fw_id)
 
+        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
+
+
     def reignite_wf(self, fw_id):
         wf = self.get_wf_by_fw_id(fw_id)
         for fw in wf.fws:
             self.reignite_fw(fw.fw_id)
 
+    def archive_wf(self, fw_id):
+        # first archive all the launches, so they are not used in duplicate checks
+        wf = self.get_wf_by_fw_id(fw_id)
+        fw_ids = [f.fw_id for f in wf.fws]
+        for fw_id in fw_ids:
+            self.rerun_fw(fw_id)
+
+        # second set the state of all FWs to ARCHIVED
+        wf = self.get_wf_by_fw_id(fw_id)
+        for fw in wf.fws:
+            self.fireworks.find_and_modify({'fw_id': fw.fw_id}, {'$set': {'state': 'ARCHIVED'}})
+
+        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
+
     def _restart_ids(self, next_fw_id, next_launch_id):
         """
         (internal method) Used to reset id counters
@@ -270,9 +359,11 @@ class LaunchPad(FWSerializable):
         :param next_launch_id: id to give next Launch (int)
         """
         self.fw_id_assigner.remove()
-        self.fw_id_assigner.find_and_modify({'_id': -1}, {'next_fw_id': next_fw_id, 'next_launch_id': next_launch_id},
+        self.fw_id_assigner.find_and_modify({'_id': -1}, {'next_fw_id': next_fw_id,
+                                                          'next_launch_id': next_launch_id},
                                             upsert=True)
-        self.m_logger.debug('RESTARTED fw_id, launch_id to ({}, {})'.format(next_fw_id, next_launch_id))
+        self.m_logger.debug(
+            'RESTARTED fw_id, launch_id to ({}, {})'.format(next_fw_id, next_launch_id))
 
     def _check_fw_for_uniqueness(self, m_fw):
         # check if there are duplicates
@@ -302,14 +393,19 @@ class LaunchPad(FWSerializable):
                 m_fw = self.fireworks.find_and_modify(m_query, {'$set': {'state': 'RESERVED'}},
                                                       sort=[("spec._priority", DESCENDING)])
             else:
-                m_fw = self.fireworks.find_one(m_query, {'fw_id': 1}, sort=[("spec._priority", DESCENDING)])
-
+                m_fw = self.fireworks.find_one(m_query, {'fw_id': 1, 'spec': 1},
+                                               sort=[("spec._priority", DESCENDING)])
+                
             if not m_fw:
                 return None
-            m_fw = self.get_fw_by_id(m_fw['fw_id'])
 
-            if self._check_fw_for_uniqueness(m_fw):
-                return m_fw
+            if not checkout and not '_dupefinder' in m_fw['spec']:
+                return True
+
+            # can't help but create the FW object...
+            fw = self.get_fw_by_id(m_fw['fw_id'])
+            if self._check_fw_for_uniqueness(fw):
+                return fw
 
     def _reserve_fw(self, fworker, launch_dir, host=None, ip=None):
         m_fw = self._get_a_fw_to_run(fworker.query)
@@ -318,7 +414,8 @@ class LaunchPad(FWSerializable):
             # create a launch
         # TODO: this code is duplicated with checkout_fw with minimal mods, should refactor this!!
         launch_id = self.get_new_launch_id()
-        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, launch_id=launch_id, fw_id=m_fw.fw_id)
+        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, launch_id=launch_id,
+                          fw_id=m_fw.fw_id)
         self._upsert_launch(m_launch)
 
         # add launch to FW
@@ -343,7 +440,8 @@ class LaunchPad(FWSerializable):
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
         bad_launch_data = self.launches.find({'state': 'RESERVED', 'state_history': {
-            '$elemMatch': {'state': 'RESERVED', 'updated_on': {'$lte': cutoff_timestr}}}}, {'launch_id': 1})
+            '$elemMatch': {'state': 'RESERVED', 'updated_on': {'$lte': cutoff_timestr}}}},
+                                             {'launch_id': 1})
         for ld in bad_launch_data:
             bad_launch_ids.append(ld['launch_id'])
         if fix:
@@ -352,7 +450,7 @@ class LaunchPad(FWSerializable):
         return bad_launch_ids
 
     def mark_fizzled(self, launch_id):
-        # TODO: this seems a lot like the code in _complete_launch...DRY
+        # TODO: this seems a lot like the code in complete_launch...DRY
 
         # Do a confirmed write and make sure state_history is preserved
         m_launch = self.get_launch_by_id(launch_id)
@@ -369,7 +467,8 @@ class LaunchPad(FWSerializable):
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
         bad_launch_data = self.launches.find({'state': 'RUNNING', 'state_history': {
-            '$elemMatch': {'state': 'RUNNING', 'updated_on': {'$lte': cutoff_timestr}}}}, {'launch_id': 1})
+            '$elemMatch': {'state': 'RUNNING', 'updated_on': {'$lte': cutoff_timestr}}}},
+                                             {'launch_id': 1})
         for ld in bad_launch_data:
             bad_launch_ids.append(ld['launch_id'])
         if fix:
@@ -382,7 +481,7 @@ class LaunchPad(FWSerializable):
         m_launch.set_reservation_id(reservation_id)
         self.launches.find_and_modify({'launch_id': launch_id}, m_launch.to_db_dict())
 
-    def _checkout_fw(self, fworker, launch_dir, fw_id=None, host=None, ip=None):
+    def checkout_fw(self, fworker, launch_dir, fw_id=None, host=None, ip=None):
         """
         (internal method) Finds a FireWork that's ready to be run, marks it as running,
         and returns it to the caller. The caller is responsible for running the FireWork.
@@ -407,7 +506,8 @@ class LaunchPad(FWSerializable):
 
         state_history = reserved_launch.state_history if reserved_launch else None
         l_id = reserved_launch.launch_id if reserved_launch else self.get_new_launch_id()
-        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, state_history=state_history, launch_id=l_id,
+        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, state_history=state_history,
+                          launch_id=l_id,
                           fw_id=m_fw.fw_id)
 
         self._upsert_launch(m_launch)
@@ -419,14 +519,16 @@ class LaunchPad(FWSerializable):
             m_fw.launches.append(m_launch)
         else:
             # we're updating an existing launch
-            m_fw.launches = [m_launch if l.launch_id == m_launch.launch_id else l for l in m_fw.launches]
+            m_fw.launches = [m_launch if l.launch_id == m_launch.launch_id else l for l in
+                             m_fw.launches]
 
         m_fw.state = 'RUNNING'
         self._upsert_fws([m_fw])
 
         # update any duplicated runs
-        for fw in self.fireworks.find({'launches': l_id, 'state': {'$in': ['WAITING', 'READY', 'RESERVED', 'FIZZLED']}},
-                                      {'fw_id': 1}):
+        for fw in self.fireworks.find(
+                {'launches': l_id, 'state': {'$in': ['WAITING', 'READY', 'RESERVED', 'FIZZLED']}},
+                {'fw_id': 1}):
             fw_id = fw['fw_id']
             fw = self.get_fw_by_id(fw_id)
             fw.state = 'RUNNING'
@@ -436,7 +538,12 @@ class LaunchPad(FWSerializable):
 
         return m_fw, l_id
 
-    def _complete_launch(self, launch_id, action, state='COMPLETED'):
+    def _change_launch_dir(self, launch_id, launch_dir):
+        m_launch = self.get_launch_by_id(launch_id)
+        m_launch.launch_dir = launch_dir
+        self._upsert_launch(m_launch)
+
+    def complete_launch(self, launch_id, action, state='COMPLETED'):
         """
         (internal method) used to mark a FireWork's Launch as completed.
         :param launch_id:
@@ -453,10 +560,11 @@ class LaunchPad(FWSerializable):
             fw_id = fw['fw_id']
             self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
 
-    def _ping_launch(self, launch_id):
+    def ping_launch(self, launch_id):
         m_launch = self.get_launch_by_id(launch_id)
         m_launch.touch_history()
-        self.launches.find_and_modify({'launch_id': launch_id}, m_launch.to_db_dict())
+        self.launches.find_and_modify({'launch_id': launch_id, 'state': 'RUNNING'},
+            {'$set':{'state_history':m_launch.to_db_dict()['state_history']}})
 
     def get_new_fw_id(self):
         """
@@ -468,7 +576,8 @@ class LaunchPad(FWSerializable):
         """
         Checkout the next Launch id
         """
-        return self.fw_id_assigner.find_and_modify({}, {'$inc': {'next_launch_id': 1}})['next_launch_id']
+        return self.fw_id_assigner.find_and_modify({}, {'$inc': {'next_launch_id': 1}})[
+            'next_launch_id']
 
     def _upsert_fws(self, fws, reassign_all=False):
         old_new = {} # mapping between old and new FireWork ids
@@ -485,6 +594,11 @@ class LaunchPad(FWSerializable):
 
         return old_new
 
+    def rerun_fw(self, fw_id):
+        wf = self.get_wf_by_fw_id(fw_id)
+        updated_ids = wf.rerun_fw(fw_id)
+        self._upsert_wf(wf, updated_ids)
+
     def _refresh_wf(self, wf, fw_id):
 
         """
@@ -496,11 +610,16 @@ class LaunchPad(FWSerializable):
         # TODO: need a try-except here, high probability of failure if incorrect action supplied
 
         updated_ids = wf.refresh(fw_id)
+        self._upsert_wf(wf, updated_ids)
+
+
+    def _upsert_wf(self, wf, updated_ids):
         updated_fws = [wf.id_fw[fid] for fid in updated_ids]
         old_new = self._upsert_fws(updated_fws)
         wf._reassign_ids(old_new)
-        # redo the links
-        self.workflows.find_and_modify({'nodes': fw_id}, wf.to_db_dict())
+        # redo the links - note that if you don't search across all keys, you can get errors
+        self.workflows.find_and_modify({'nodes': {'$in': wf.id_fw.keys()}}, wf.to_db_dict())
+
 
     def _steal_launches(self, thief_fw):
         stolen = False
@@ -513,43 +632,29 @@ class LaunchPad(FWSerializable):
             self.m_logger.debug('Querying for duplicates, fw_id: {}'.format(thief_fw.fw_id))
             for potential_match in self.fireworks.find(m_query):
                 self.m_logger.debug(
-                    'Verifying for duplicates, fw_ids: {}, {}'.format(thief_fw.fw_id, potential_match['fw_id']))
+                    'Verifying for duplicates, fw_ids: {}, {}'.format(thief_fw.fw_id,
+                                                                      potential_match['fw_id']))
                 spec1 = dict(thief_fw.to_dict()['spec'])  # defensive copy
                 spec2 = dict(potential_match['spec'])  # defensive copy
                 if m_dupefinder.verify(spec1, spec2):  # verify the match
                     # steal the launches
                     victim_fw = self.get_fw_by_id(potential_match['fw_id'])
                     thief_launches = [l.launch_id for l in thief_fw.launches]
-                    valuable_launches = [l for l in victim_fw.launches if l.launch_id not in thief_launches]
+                    valuable_launches = [l for l in victim_fw.launches if
+                                         l.launch_id not in thief_launches]
                     for launch in valuable_launches:
                         thief_fw.launches.append(launch)
                         stolen = True
                         self.m_logger.info(
-                            'Duplicate found! fwids {} and {}'.format(thief_fw.fw_id, potential_match['fw_id']))
+                            'Duplicate found! fwids {} and {}'.format(thief_fw.fw_id,
+                                                                      potential_match['fw_id']))
         return stolen
 
     def _upsert_launch(self, m_launch):
         # Do a confirmed write of Launch
-
+        # TODO: this no longer needs to be its own function (much was removed)
         self.launches.find_and_modify({'launch_id': m_launch.launch_id}, m_launch.to_db_dict(), upsert=True)
 
-        l_id = m_launch.launch_id
-        # confirm write
-        # I can't believe this is actually necessary (and yes, it appears to be necessary)
-        nloops = 0
-        while not self.launches.find_one({'launch_id': l_id, 'state': m_launch.state}):
-            self.m_logger.debug('Waiting for a delayed write of Launch...')
-            nloops += 1
-            if nloops == 80:
-                if m_launch.state == 'FIZZLED':
-                    # We're unable to mark the launch as FIZZLED
-                    self.m_logger.critical('UNABLE to mark launch_id: {} as fizzled!'.format(l_id))
-                    break
-                # mark the launch as FIZZLED because we're unable to update its state...
-                self.m_logger.error('FIZZLED launch id: {} because could not confirm write!!'.format(l_id))
-                self.mark_fizzled(l_id)
-                break
-            if nloops % 20 == 0:
-                self.m_logger.info('Fixing a lost write of Launch!!')
-                self.launches.find_and_modify({'launch_id': l_id}, m_launch.to_db_dict())
-            time.sleep(4)
\ No newline at end of file
+    def get_logdir(self):
+        # AJ: This is needed for job packing due to Proxy objects not being fully featured...
+        return self.logdir
\ No newline at end of file
